__version__ = "0.0.17"
# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/00__init__.ipynb.

# %% auto 0
__all__ = []

# %% ../nbs/00__init__.ipynb 3
from itertools import chain
from . import (cons, duck, atyp, grds, util, prot,)

from .cons import *
from .duck import *
from .atyp import *
from .grds import *
from .util import *
from .prot import *

# %% ../nbs/00__init__.ipynb 4
__all__ = sorted(set(chain(*[
    mod.__all__ for mod in (cons, duck, atyp, grds, util, prot,)
])))



# Autogenerated by nbdev

d = { 'settings': { 'branch': 'main',
                'doc_baseurl': '/ptyp',
                'doc_host': 'https://dsm-72.github.io',
                'git_url': 'https://github.com/dsm-72/ptyp',
                'lib_path': 'ptyp'},
  'syms': { 'ptyp.atyp': {},
            'ptyp.cons': { 'ptyp.cons.ProtocolMethod': ('cons.html#protocolmethod', 'ptyp/cons.py'),
                           'ptyp.cons.ProtocolMethod._missing_': ('cons.html#protocolmethod._missing_', 'ptyp/cons.py'),
                           'ptyp.cons.ProtocolMethod.casefold': ('cons.html#protocolmethod.casefold', 'ptyp/cons.py'),
                           'ptyp.cons.ProtocolMethod.gidt': ('cons.html#protocolmethod.gidt', 'ptyp/cons.py'),
                           'ptyp.cons.ProtocolMethod.idgt': ('cons.html#protocolmethod.idgt', 'ptyp/cons.py'),
                           'ptyp.cons.ProtocolMethod.ieq': ('cons.html#protocolmethod.ieq', 'ptyp/cons.py'),
                           'ptyp.cons.ProtocolMethod.startswith': ('cons.html#protocolmethod.startswith', 'ptyp/cons.py'),
                           'ptyp.cons.ProtocolMethodOrder': ('cons.html#protocolmethodorder', 'ptyp/cons.py'),
                           'ptyp.cons.synonyms': ('cons.html#synonyms', 'ptyp/cons.py'),
                           'ptyp.cons.synonyms.__eq__': ('cons.html#synonyms.__eq__', 'ptyp/cons.py'),
                           'ptyp.cons.synonyms.__hash__': ('cons.html#synonyms.__hash__', 'ptyp/cons.py'),
                           'ptyp.cons.synonyms.__new__': ('cons.html#synonyms.__new__', 'ptyp/cons.py'),
                           'ptyp.cons.synonyms.__repr__': ('cons.html#synonyms.__repr__', 'ptyp/cons.py'),
                           'ptyp.cons.synonyms.__str__': ('cons.html#synonyms.__str__', 'ptyp/cons.py'),
                           'ptyp.cons.synonyms.startswith': ('cons.html#synonyms.startswith', 'ptyp/cons.py'),
                           'ptyp.cons.synonyms.strs': ('cons.html#synonyms.strs', 'ptyp/cons.py')},
            'ptyp.duck': { 'ptyp.duck.Hashable': ('duck.html#hashable', 'ptyp/duck.py'),
                           'ptyp.duck.Hashable.__hash__': ('duck.html#hashable.__hash__', 'ptyp/duck.py'),
                           'ptyp.duck.Hashable.__instancecheck__': ('duck.html#hashable.__instancecheck__', 'ptyp/duck.py'),
                           'ptyp.duck.NonStrBytesIterable': ('duck.html#nonstrbytesiterable', 'ptyp/duck.py'),
                           'ptyp.duck.NonStrBytesIterable.__instancecheck__': ( 'duck.html#nonstrbytesiterable.__instancecheck__',
                                                                                'ptyp/duck.py'),
                           'ptyp.duck.NonStrBytesIterable.__iter__': ('duck.html#nonstrbytesiterable.__iter__', 'ptyp/duck.py'),
                           'ptyp.duck.NotSpecial': ('duck.html#notspecial', 'ptyp/duck.py'),
                           'ptyp.duck.NotSpecial.__hash__': ('duck.html#notspecial.__hash__', 'ptyp/duck.py'),
                           'ptyp.duck.NotSpecial.__instancecheck__': ('duck.html#notspecial.__instancecheck__', 'ptyp/duck.py'),
                           'ptyp.duck._acczeros': ('duck.html#_acczeros', 'ptyp/duck.py'),
                           'ptyp.duck._format_list': ('duck.html#_format_list', 'ptyp/duck.py'),
                           'ptyp.duck._genzeros': ('duck.html#_genzeros', 'ptyp/duck.py'),
                           'ptyp.duck._shape': ('duck.html#_shape', 'ptyp/duck.py'),
                           'ptyp.duck.empty': ('duck.html#empty', 'ptyp/duck.py'),
                           'ptyp.duck.empty.__init__': ('duck.html#empty.__init__', 'ptyp/duck.py'),
                           'ptyp.duck.empty.__repr__': ('duck.html#empty.__repr__', 'ptyp/duck.py'),
                           'ptyp.duck.empty.ndim': ('duck.html#empty.ndim', 'ptyp/duck.py')},
            'ptyp.grds': { 'ptyp.grds.isalias': ('grds.html#isalias', 'ptyp/grds.py'),
                           'ptyp.grds.isallsame': ('grds.html#isallsame', 'ptyp/grds.py'),
                           'ptyp.grds.isargs': ('grds.html#isargs', 'ptyp/grds.py'),
                           'ptyp.grds.iscls': ('grds.html#iscls', 'ptyp/grds.py'),
                           'ptyp.grds.isdict': ('grds.html#isdict', 'ptyp/grds.py'),
                           'ptyp.grds.isdunder': ('grds.html#isdunder', 'ptyp/grds.py'),
                           'ptyp.grds.ishashable': ('grds.html#ishashable', 'ptyp/grds.py'),
                           'ptyp.grds.isiter': ('grds.html#isiter', 'ptyp/grds.py'),
                           'ptyp.grds.isnan': ('grds.html#isnan', 'ptyp/grds.py'),
                           'ptyp.grds.isnotiterstr': ('grds.html#isnotiterstr', 'ptyp/grds.py'),
                           'ptyp.grds.isnotspecial': ('grds.html#isnotspecial', 'ptyp/grds.py'),
                           'ptyp.grds.isorigin': ('grds.html#isorigin', 'ptyp/grds.py'),
                           'ptyp.grds.isprivate': ('grds.html#isprivate', 'ptyp/grds.py'),
                           'ptyp.grds.isstr': ('grds.html#isstr', 'ptyp/grds.py'),
                           'ptyp.grds.istuple': ('grds.html#istuple', 'ptyp/grds.py'),
                           'ptyp.grds.istype': ('grds.html#istype', 'ptyp/grds.py'),
                           'ptyp.grds.istyping': ('grds.html#istyping', 'ptyp/grds.py'),
                           'ptyp.grds.notcls': ('grds.html#notcls', 'ptyp/grds.py')},
            'ptyp.meta': { 'ptyp.meta.Alias': ('meta.html#alias', 'ptyp/meta.py'),
                           'ptyp.meta.AliasMeta': ('meta.html#aliasmeta', 'ptyp/meta.py'),
                           'ptyp.meta.AliasMeta.getbtype': ('meta.html#aliasmeta.getbtype', 'ptyp/meta.py'),
                           'ptyp.meta.Not': ('meta.html#not', 'ptyp/meta.py'),
                           'ptyp.meta.NotMeta': ('meta.html#notmeta', 'ptyp/meta.py'),
                           'ptyp.meta.NotMeta.check': ('meta.html#notmeta.check', 'ptyp/meta.py'),
                           'ptyp.meta.NotMeta.default_kwds': ('meta.html#notmeta.default_kwds', 'ptyp/meta.py'),
                           'ptyp.meta.NotMeta.getbtype': ('meta.html#notmeta.getbtype', 'ptyp/meta.py'),
                           'ptyp.meta.Opt': ('meta.html#opt', 'ptyp/meta.py'),
                           'ptyp.meta.OptionalMeta': ('meta.html#optionalmeta', 'ptyp/meta.py'),
                           'ptyp.meta.OptionalMeta.__new__': ('meta.html#optionalmeta.__new__', 'ptyp/meta.py'),
                           'ptyp.meta.OptionalMeta.default_kwds': ('meta.html#optionalmeta.default_kwds', 'ptyp/meta.py'),
                           'ptyp.meta.OptionalMeta.getbtype': ('meta.html#optionalmeta.getbtype', 'ptyp/meta.py'),
                           'ptyp.meta.ProtoType': ('meta.html#prototype', 'ptyp/meta.py'),
                           'ptyp.meta.ProtocolBaseMeta': ('meta.html#protocolbasemeta', 'ptyp/meta.py'),
                           'ptyp.meta.ProtocolBaseMeta.__call__': ('meta.html#protocolbasemeta.__call__', 'ptyp/meta.py'),
                           'ptyp.meta.ProtocolBaseMeta.__init_subclass__': ('meta.html#protocolbasemeta.__init_subclass__', 'ptyp/meta.py'),
                           'ptyp.meta.ProtocolBaseMeta.__instancecheck__': ('meta.html#protocolbasemeta.__instancecheck__', 'ptyp/meta.py'),
                           'ptyp.meta.ProtocolBaseMeta.__new__': ('meta.html#protocolbasemeta.__new__', 'ptyp/meta.py'),
                           'ptyp.meta.ProtocolBaseMeta.check': ('meta.html#protocolbasemeta.check', 'ptyp/meta.py'),
                           'ptyp.meta.ProtocolBaseMeta.classname': ('meta.html#protocolbasemeta.classname', 'ptyp/meta.py'),
                           'ptyp.meta.ProtocolBaseMeta.default_kwds': ('meta.html#protocolbasemeta.default_kwds', 'ptyp/meta.py'),
                           'ptyp.meta.ProtocolBaseMeta.getallattrs': ('meta.html#protocolbasemeta.getallattrs', 'ptyp/meta.py'),
                           'ptyp.meta.ProtocolBaseMeta.getattrs': ('meta.html#protocolbasemeta.getattrs', 'ptyp/meta.py'),
                           'ptyp.meta.ProtocolBaseMeta.getbtype': ('meta.html#protocolbasemeta.getbtype', 'ptyp/meta.py'),
                           'ptyp.meta.ProtocolBaseMeta.getguards': ('meta.html#protocolbasemeta.getguards', 'ptyp/meta.py'),
                           'ptyp.meta.ProtocolBaseMeta.gethasattrs': ('meta.html#protocolbasemeta.gethasattrs', 'ptyp/meta.py'),
                           'ptyp.meta.ProtocolBaseMeta.getmapattrs': ('meta.html#protocolbasemeta.getmapattrs', 'ptyp/meta.py'),
                           'ptyp.meta.ProtocolBaseMeta.getqualname': ('meta.html#protocolbasemeta.getqualname', 'ptyp/meta.py'),
                           'ptyp.meta.ProtocolBaseMeta.gettypes': ('meta.html#protocolbasemeta.gettypes', 'ptyp/meta.py'),
                           'ptyp.meta.ProtocolBaseMeta.makebases': ('meta.html#protocolbasemeta.makebases', 'ptyp/meta.py'),
                           'ptyp.meta.ProtocolBaseMeta.makeguard': ('meta.html#protocolbasemeta.makeguard', 'ptyp/meta.py'),
                           'ptyp.meta.ProtocolBaseMeta.metadict': ('meta.html#protocolbasemeta.metadict', 'ptyp/meta.py'),
                           'ptyp.meta.ProtocolBaseMeta.qualname': ('meta.html#protocolbasemeta.qualname', 'ptyp/meta.py'),
                           'ptyp.meta.ProtocolBaseMeta.setallattrs': ('meta.html#protocolbasemeta.setallattrs', 'ptyp/meta.py'),
                           'ptyp.meta.ProtocolBaseMeta.setattrs': ('meta.html#protocolbasemeta.setattrs', 'ptyp/meta.py'),
                           'ptyp.meta.ProtocolBaseMeta.setbtype': ('meta.html#protocolbasemeta.setbtype', 'ptyp/meta.py'),
                           'ptyp.meta.ProtocolBaseMeta.setguards': ('meta.html#protocolbasemeta.setguards', 'ptyp/meta.py'),
                           'ptyp.meta.ProtocolBaseMeta.sethasattrs': ('meta.html#protocolbasemeta.sethasattrs', 'ptyp/meta.py'),
                           'ptyp.meta.ProtocolBaseMeta.setmapattrs': ('meta.html#protocolbasemeta.setmapattrs', 'ptyp/meta.py'),
                           'ptyp.meta.ProtocolBaseMeta.setqualname': ('meta.html#protocolbasemeta.setqualname', 'ptyp/meta.py'),
                           'ptyp.meta.ProtocolBaseMeta.settypes': ('meta.html#protocolbasemeta.settypes', 'ptyp/meta.py'),
                           'ptyp.meta.ProtocolMeta': ('meta.html#protocolmeta', 'ptyp/meta.py'),
                           'ptyp.meta.ProtocolMeta.getbtype': ('meta.html#protocolmeta.getbtype', 'ptyp/meta.py')},
            'ptyp.prot': { 'ptyp.prot.Alias': ('prot.html#alias', 'ptyp/prot.py'),
                           'ptyp.prot.AliasMeta': ('prot.html#aliasmeta', 'ptyp/prot.py'),
                           'ptyp.prot.AliasMeta.getbtype': ('prot.html#aliasmeta.getbtype', 'ptyp/prot.py'),
                           'ptyp.prot.Not': ('prot.html#not', 'ptyp/prot.py'),
                           'ptyp.prot.NotMeta': ('prot.html#notmeta', 'ptyp/prot.py'),
                           'ptyp.prot.NotMeta.default_kwds': ('prot.html#notmeta.default_kwds', 'ptyp/prot.py'),
                           'ptyp.prot.NotMeta.getbtype': ('prot.html#notmeta.getbtype', 'ptyp/prot.py'),
                           'ptyp.prot.NotMeta.guard': ('prot.html#notmeta.guard', 'ptyp/prot.py'),
                           'ptyp.prot.Opt': ('prot.html#opt', 'ptyp/prot.py'),
                           'ptyp.prot.OptionalMeta': ('prot.html#optionalmeta', 'ptyp/prot.py'),
                           'ptyp.prot.OptionalMeta.__new__': ('prot.html#optionalmeta.__new__', 'ptyp/prot.py'),
                           'ptyp.prot.OptionalMeta.default_kwds': ('prot.html#optionalmeta.default_kwds', 'ptyp/prot.py'),
                           'ptyp.prot.OptionalMeta.getbtype': ('prot.html#optionalmeta.getbtype', 'ptyp/prot.py'),
                           'ptyp.prot.ProtocolBaseMeta': ('prot.html#protocolbasemeta', 'ptyp/prot.py'),
                           'ptyp.prot.ProtocolBaseMeta.__call__': ('prot.html#protocolbasemeta.__call__', 'ptyp/prot.py'),
                           'ptyp.prot.ProtocolBaseMeta.__init__': ('prot.html#protocolbasemeta.__init__', 'ptyp/prot.py'),
                           'ptyp.prot.ProtocolBaseMeta.__init_subclass__': ('prot.html#protocolbasemeta.__init_subclass__', 'ptyp/prot.py'),
                           'ptyp.prot.ProtocolBaseMeta.__instancecheck__': ('prot.html#protocolbasemeta.__instancecheck__', 'ptyp/prot.py'),
                           'ptyp.prot.ProtocolBaseMeta.__instanceguard__': ('prot.html#protocolbasemeta.__instanceguard__', 'ptyp/prot.py'),
                           'ptyp.prot.ProtocolBaseMeta.__invert__': ('prot.html#protocolbasemeta.__invert__', 'ptyp/prot.py'),
                           'ptyp.prot.ProtocolBaseMeta.__invopt__': ('prot.html#protocolbasemeta.__invopt__', 'ptyp/prot.py'),
                           'ptyp.prot.ProtocolBaseMeta.__new__': ('prot.html#protocolbasemeta.__new__', 'ptyp/prot.py'),
                           'ptyp.prot.ProtocolBaseMeta.__not__': ('prot.html#protocolbasemeta.__not__', 'ptyp/prot.py'),
                           'ptyp.prot.ProtocolBaseMeta.__redirect__': ('prot.html#protocolbasemeta.__redirect__', 'ptyp/prot.py'),
                           'ptyp.prot.ProtocolBaseMeta.check': ('prot.html#protocolbasemeta.check', 'ptyp/prot.py'),
                           'ptyp.prot.ProtocolBaseMeta.construct': ('prot.html#protocolbasemeta.construct', 'ptyp/prot.py'),
                           'ptyp.prot.ProtocolBaseMeta.default_kwds': ('prot.html#protocolbasemeta.default_kwds', 'ptyp/prot.py'),
                           'ptyp.prot.ProtocolBaseMeta.getbtype': ('prot.html#protocolbasemeta.getbtype', 'ptyp/prot.py'),
                           'ptyp.prot.ProtocolBaseMeta.guard': ('prot.html#protocolbasemeta.guard', 'ptyp/prot.py'),
                           'ptyp.prot.ProtocolBaseMeta.makebases': ('prot.html#protocolbasemeta.makebases', 'ptyp/prot.py'),
                           'ptyp.prot.ProtocolBaseMeta.new': ('prot.html#protocolbasemeta.new', 'ptyp/prot.py'),
                           'ptyp.prot.ProtocolBaseMeta.setqualname': ('prot.html#protocolbasemeta.setqualname', 'ptyp/prot.py'),
                           'ptyp.prot.ProtocolBaseMeta.typeguard': ('prot.html#protocolbasemeta.typeguard', 'ptyp/prot.py'),
                           'ptyp.prot.ProtocolMeta': ('prot.html#protocolmeta', 'ptyp/prot.py'),
                           'ptyp.prot.ProtocolMeta.getbtype': ('prot.html#protocolmeta.getbtype', 'ptyp/prot.py'),
                           'ptyp.prot.Prototype': ('prot.html#prototype', 'ptyp/prot.py'),
                           'ptyp.prot.PrototypeMeta': ('prot.html#prototypemeta', 'ptyp/prot.py'),
                           'ptyp.prot.PrototypeMeta.__new__': ('prot.html#prototypemeta.__new__', 'ptyp/prot.py'),
                           'ptyp.prot.PrototypeMeta.getbtype': ('prot.html#prototypemeta.getbtype', 'ptyp/prot.py')},
            'ptyp.ptyp': { 'ptyp.ptyp.IntQ': ('ptyp.html#intq', 'ptyp/ptyp.py'),
                           'ptyp.ptyp.NotStrBytes': ('ptyp.html#notstrbytes', 'ptyp/ptyp.py'),
                           'ptyp.ptyp.StrBytes': ('ptyp.html#strbytes', 'ptyp/ptyp.py'),
                           'ptyp.ptyp.StrQ': ('ptyp.html#strq', 'ptyp/ptyp.py')},
            'ptyp.type': {},
            'ptyp.util': { 'ptyp.util.argname': ('util.html#argname', 'ptyp/util.py'),
                           'ptyp.util.argsname': ('util.html#argsname', 'ptyp/util.py'),
                           'ptyp.util.classname': ('util.html#classname', 'ptyp/util.py'),
                           'ptyp.util.filtkwds': ('util.html#filtkwds', 'ptyp/util.py'),
                           'ptyp.util.getname': ('util.html#getname', 'ptyp/util.py'),
                           'ptyp.util.getorigintypes': ('util.html#getorigintypes', 'ptyp/util.py'),
                           'ptyp.util.guardtype': ('util.html#guardtype', 'ptyp/util.py'),
                           'ptyp.util.opttypes': ('util.html#opttypes', 'ptyp/util.py'),
                           'ptyp.util.qualname': ('util.html#qualname', 'ptyp/util.py'),
                           'ptyp.util.unionify': ('util.html#unionify', 'ptyp/util.py')}}}



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/05_util.ipynb.

# %% auto 0
__all__ = ['unionify', 'getorigintypes', 'opttypes', 'guardtype', 'filtkwds', 'getname', 'argname', 'argsname', 'classname',
           'qualname']

# %% ../nbs/05_util.ipynb 6
import inspect
from inspect import signature

# %% ../nbs/05_util.ipynb 8
import typing
from typing import (Any, Dict, Union, TypeGuard, Callable, Optional, get_args, )

# %% ../nbs/05_util.ipynb 10
#| export


# %% ../nbs/05_util.ipynb 12
from nchr import NIL
from nlit import __NAME__

# %% ../nbs/05_util.ipynb 14
from .cons import dunname
from .duck import NotSpecial
from .atyp import T, Types, Guard
from .grds import isnotspecial

# %% ../nbs/05_util.ipynb 16
def unionify(*t: Types) -> Union[None, type, Types]:
    '''Create a union type from given types, handling `None` and single type cases.'''
    return None if (n := len(t)) == 0 else t[0] if n == 1 else Union[t[0], *t[1:]]

def getorigintypes() -> list[NotSpecial]:
    '''Retrieve a list of non-special types from the `typing` module.'''
    return [a for e in dir(typing) if isnotspecial(a := getattr(typing, e))]

def opttypes(types: Types, dropnone: bool = True) -> Types:
    '''Optionally wrap types in a `typing.Optional` type, excluding `None` if specified.

    Parameters
    ----------
    types : Types
        A tuple of types to be optionally wrapped in `typing.Optional`.
        
    dropnone : bool, default True
        Whether to exclude `None` type from the resulting tuple.

    Returns
    -------
    Types
        A tuple of types, optionally wrapped in `typing.Optional`.
    '''
    return types if not dropnone else tuple(Optional[t] for t in types if t is not None)

def guardtype(*types: Types) -> Guard:
    '''Create a TypeGuard for a union of the given types.

    Parameters
    ----------
    *types : Types
        A variadic number of type variables to be included in the class's qualified name.
        
    Returns
    -------
    Guard
        A TypeGuard for the union of the given types.
    '''
    return TypeGuard[unionify(*opttypes(types))]

# %% ../nbs/05_util.ipynb 18
def filtkwds(fn: Callable, **kwargs) -> Dict[str, Any]:
    '''Filter keyword arguments to only include those that are valid for a given function's signature.

    Parameters
    ----------
    fn : Callable
        The function whose signature is to be considered for filtering.
        
    **kwargs : dict
        Keyword arguments to be filtered based on the function's signature.

    Returns
    -------
    Dict[str, Any]
        A dictionary of keyword arguments that are valid for the function's signature.
    '''
    sig = signature(fn).parameters
    return {k: v for k, v in kwargs.items() if k in sig}

def getname(o: object) -> str:
    '''Get the name of an object or its string representation if the name is not directly accessible.

    Parameters
    ----------
    o : object
        The object whose name is to be retrieved.

    Returns
    -------
    str
        The name of the object or its string representation.
    '''
    return str(dunname(o))

def argname(a: T) -> str:
    '''Retrieve the name of the first argument type for a given type variable.

    Parameters
    ----------
    a : T
        A type variable whose first argument type's name is to be retrieved.

    Returns
    -------
    str
        The name of the first argument type of the type variable.
    '''
    return getname((get_args(a) or (a, ))[0])

def argsname(*args: Types, sep: str = ', ') -> str:
    '''Concatenate the names of the argument types for a list of type variables.

    Parameters
    ----------
    *args : Types
        A variadic number of type variables.
        
    sep : str, default: ', '
        The separator to use when concatenating the names of the argument types.

    Returns
    -------
    str
        A comma-separated string of the names of the argument types.
    '''
    return sep.join(list(map(argname, args)))

def classname(cls: type, alt: Optional[str] = None) -> str:
    '''Retrieve the name of a class, or an alternate name if provided.

    Parameters
    ----------
    cls : type
        The class whose name is to be retrieved.
        
    alt : Optional[str], optional
        An alternate name to return instead of the class's name.

    Returns
    -------
    str
        The name of the class or the alternate name if provided.
    '''
    if inspect.isclass(alt): return getname(alt)
    if isinstance(alt, str): return alt
    return getname(cls)

def qualname(
    cls: type, 
    *types: Types, 
    prefix: str = NIL, 
    suffix: str = NIL, 
    dropnone: bool = True, 
    alt: Optional[str] = None,
) -> str:
    '''Generate a qualified name for a class with specific type parameters and optional prefix/suffix.

    Parameters
    ----------
    cls : type
        The class for which the qualified name is to be generated.
        
    *types : Types
        A variadic number of type variables to be included in the class's qualified name.
        
    prefix : str, optional
        A prefix to be added to the qualified name.
        
    suffix : str, optional
        A suffix to be added to the qualified name.
        
    dropnone : bool, default True
        Whether to exclude `None` type from the type variables.
        
    alt : Optional[str], optional
        An alternative name to use instead of the class's name.

    Returns
    -------
    str
        The generated qualified name of the class.
    '''
    types = opttypes(types, dropnone)
    cname = classname(cls, alt)
    qname = f'{prefix}{{{argsname(*types)}}}{suffix}'    
    qname = f'{cname}({qname})'
    return qname



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/04_grds.ipynb.

# %% auto 0
__all__ = ['iscls', 'notcls', 'istuple', 'isstr', 'isdict', 'isiter', 'isnan', 'ishashable', 'isnotspecial', 'isorigin', 'isargs',
           'istyping', 'isprivate', 'isdunder', 'isnotiterstr', 'istype', 'isallsame', 'isalias']

# %% ../nbs/04_grds.ipynb 6
import io, math, inspect
from operator import eq, is_
from contextlib import redirect_stderr

# %% ../nbs/04_grds.ipynb 8
from typing import (
    TypeGuard, Iterable, get_args, get_origin,
    _SpecialForm, Generic, Protocol
)

# %% ../nbs/04_grds.ipynb 10
#| export


# %% ../nbs/04_grds.ipynb 12
from nlit import __HASH__ 
from uscr import private, dunder

# %% ../nbs/04_grds.ipynb 14
from .cons import ALLATTRS
from ptyp.atyp import (
    T, Types, IterStr, GuardIterT, AttrGuards, GuardIter, IterFunc, IterType, 
    GuardIterType, NanGuard, HasHash, IsNotSpecial, OriginGuard, Guard, OriginTypes, HasIterElements
)

# %% ../nbs/04_grds.ipynb 17
def iscls(x) -> TypeGuard[type]:
    '''Check if `x` is a class.
    
    Notes
    ----- 
    `iscls` is an alias for `inspect.isclass`.
    Class objects provide these attributes:
    
    `__doc__`: 
        documentation string 
        
    `__module__`: 
        name of module in which this class was defined
    
    See Also
    --------
    inspect.isclass : https://docs.python.org/3/library/inspect.html#inspect.isclass
    '''
    return inspect.isclass(x)

def notcls(x) -> TypeGuard[T]:
    '''Check if `x` is not a class.
    
    See Also
    --------
    iscls : Alias for `inspect.isclass`
    inspect.isclass : https://docs.python.org/3/library/inspect.html#inspect.isclass
    '''
    return not iscls(x)

def istuple(o: object) -> TypeGuard[tuple]:
    '''Check if the given object is `tuple`.'''
    return isinstance(o, tuple)

def isstr(x) -> TypeGuard[str]:
    '''Check if `x` is an `str`.'''
    return isinstance(x, str)

def isdict(x) -> TypeGuard[dict]:
    '''Check if `x` is a `dict`.'''
    return isinstance(x, dict)

def isiter(o: object) -> GuardIter: 
    '''Check if the given object is an iterable.'''
    return isinstance(o, Iterable)

def isnan(o: object) -> NanGuard:
    '''Check if the given object is `math.nan`.'''
    try: return math.isnan(o)
    except: return False

# %% ../nbs/04_grds.ipynb 19
def ishashable(o: object) -> HasHash:
    '''Check if the given object is hashable.'''
    return hasattr(o, __HASH__)

def isnotspecial(t: type) -> IsNotSpecial:
    '''Check if the given type is not a special form in the `typing` module.'''
    withorg = get_origin(t)
    hashash = ishashable(t)
    special = isinstance(t, _SpecialForm) or t in (Generic, Protocol)
    return withorg and hashash and not special

def isorigin(t: type) -> OriginGuard:
    '''Check if the given type is one of the origin types defined in `OriginTypes`.'''
    otypes = get_args(OriginTypes)
    return any((get_origin(t) is _ for _ in otypes))

def isargs(o: object, t: T) -> Guard:
    '''Check if the given object is equal to or contained within the arguments of the type `t`.'''
    try: 
        args = get_args(t)
        # NOTE: len(args) and o == args makes sure empty tuple() is not True
        return any(((len(args) and o == args), o in args, o == t))
    except: return False

def istyping(o: object, t: T) -> Guard:
    '''Check if the given object is a typing construct that matches type `t`.'''
    if not isorigin(t): return False
    try: return isargs(o, t)
    except: return False

# %% ../nbs/04_grds.ipynb 21
def isprivate(s: str) -> TypeGuard[private]:
    '''Check if `s` is a private string e.g. `'_{anything}'`.'''
    return isinstance(s, private)

def isdunder(s: str) -> TypeGuard[dunder]:
    '''Check if `s` is an dunder string e.g. `'__{anything}__'`.'''
    return isinstance(s, dunder)

def isnotiterstr(o: object) -> HasIterElements:
    '''Check if the given object is an iterable but not a string or bytes.'''
    return isiter(o) and not isinstance(o, (str, bytes))

# %% ../nbs/04_grds.ipynb 23
def istype(o: object, t: T, guards: IterFunc = tuple(), dropfn: IterFunc = tuple()) -> Guard:
    '''Check if the given object is of type `t`, using additional guard functions.

    Parameters
    ----------
    o : object
        The object to check.
        
    t : T
        The type to check against.
        
    guards : IterFunc, optional
        Additional guard functions to use in the check.
        
    dropfn : IterFunc, optional
        Functions to exclude from the check.

    Returns
    -------
    Guard
        True if the object is of type `t` or satisfies any of the guard functions, otherwise False.
        
    Notes
    -----
    The `guards` and `dropfn` parameters are used to filter the set of functions used in the check.
    
    The mandatory guard functions are `isinstance`, `issubclass`, `typs.istyping`, `typs.isargs`, 
        `operator.eq`, and `operator.is_`.
    '''
    funcs = (isinstance, issubclass, istyping, isargs, eq, is_,) + tuple(guards)
    funcs = (f for f in funcs if f not in dropfn)
    with redirect_stderr(io.StringIO()):
        for comp in funcs:
            try:
                if comp(o, t):
                    # print(f'{comp.__name__}({o}, {t}), True')
                    return True
            except: 
                continue
            
        if isnan(t): # checknan(o, t) ==> isnan(t) and isnan(o)
            if isnan(o): 
                return True
    # print(f'istype({o}, {t}), False')
    return False

# %% ../nbs/04_grds.ipynb 25
def isallsame(it: IterType, dtype: T) -> GuardIterType:
    '''Check if all elements in an iterable are of the same specified type.

    Parameters
    ----------
    it : IterType
        The iterable to check.
        
    dtype : T
        The type against which to check the elements of the iterable.

    Returns
    -------
    GuardIterType
        True if all elements are of the specified type, otherwise False.
    '''
    if not isiter(it): 
        return istype(it, dtype)
    
    for el in it:
        if isnotiterstr(el):
            # el is also iterable (but not like 'abc'), so recurse over the sub-iterable elements
            try:
                if not isallsame(el, dtype): 
                    return False
            except TypeError: 
                return False
            
        elif not istype(el, dtype):
            return False
    return True

# %% ../nbs/04_grds.ipynb 27
def isalias(
    obj: T,       # object to check
    types: Types, # types to check
    guards: IterFunc = tuple(),
    attrs: dict = dict(),
    hasattrs: IterStr = tuple(),  # attributes to check
    allattrs: IterStr = ALLATTRS, # attributes to check that values are all the same if object is iterable
    mapattrs: AttrGuards = dict(),
) -> GuardIterT:
    '''Check if an object matches a set of types and specific attributes.

    Parameters
    ----------
    obj : T
        The object to check.
        
    types : Types
        A tuple of types to compare against the object.
        
    guards : IterFunc, optional
        Iterable of guard functions.
        
    attrs : dict, optional
        Dictionary of attributes and their expected values to check in the object.
        
    hasattrs : IterStr, optional
        Iterable of attribute names that the object must have.
        
    allattrs : IterStr, optional
        Iterable of attributes to check if their values are all the same in an iterable object.
        
    mapattrs : AttrGuards, optional
        Mapping of attribute names to guard functions for additional checks.

    Returns
    -------
    GuardIterT
        Returns True if the object matches the specified conditions, False otherwise.
    '''
    hasattrs = tuple(hasattrs) + tuple(k for k, v in attrs.items() if k not in allattrs)
    if not all(hasattr(obj, attr) for attr in hasattrs): return False
    
    found = any(istype(obj, t, guards) for t in types)
    if not found: return False
    
    for attr, val in attrs.items():
        if not (getattr(obj, attr, None) == val):
            # return False
            continue
        
        if attr in allattrs and not (isnotiterstr(obj) and isallsame(obj, val)): 
            # return False
            continue
        
        if attr in mapattrs and not (mapattrs[attr](obj, val)):            
            # return False
            continue
        
    return True



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/01_cons.ipynb.

# %% auto 0
__all__ = ['dunname', 'ALLATTRS', 'PROTOCOL_METHOD_ORDER', 'CALLFNCS', 'ProtocolMethod', 'ProtocolMethodOrder']

# %% ../nbs/01_cons.ipynb 6
from enum import Enum

# %% ../nbs/01_cons.ipynb 8
#| export


# %% ../nbs/01_cons.ipynb 10
#| export


# %% ../nbs/01_cons.ipynb 12
from nchr import NIL
from nlit import STRS, TYPE, TYPES, DTYPE, VALUE, __NAME__, _DEFAULT_
from astr import astr

# %% ../nbs/01_cons.ipynb 14
#| export


# %% ../nbs/01_cons.ipynb 16
dunname = astr(__NAME__, retself=True)
'''Getter for the dunder `__name__` attribute of the given object.''';

ALLATTRS = {TYPE, DTYPE, TYPES}
'''Attributes of which all values in the iterable are to be checked against the type guards.''';

# %% ../nbs/01_cons.ipynb 18
class synonyms(str):
    def __new__(cls, value, *strs: str):
        new = super().__new__(cls, value)
        new._orig = str(value)
        new._strs = (getattr(value, STRS, set()) | set(strs)) - {new._orig, getattr(value, '_orig', '')}
        return new
    
    def __eq__(self, other):
        return super().__eq__(other) or str(other).casefold() in map(str.casefold, self.strs)
    
    
    def startswith(self, __prefix: str | tuple[str, ...]) -> bool:
        return any(str(s).casefold().startswith(str(__prefix).casefold()) for s in self.strs)
    
    
    @property
    def strs(self) -> set[str]:
        return {str(self)} | self._strs
    
    def __hash__(self):
        return super().__hash__()
    
    def __str__(self):
        return str.__str__(self)
    
    def __repr__(self) -> str:
        strs = str(sorted(self._strs)).replace('[', '{').replace(']', '}')
        return f'{self._orig} | {strs}'

# %% ../nbs/01_cons.ipynb 20
# Option 1: try to construct a class from the protocol's `types`
gen = synonyms('gen', 'generator', 'con', 'construct')
'''Try to generate a class from the protocol's `types`.''';

# Option 2: decorate a class
dec = synonyms('dec', 'decorator', 'dec')
'''Use the protocol as a decorator for a class.''';

# Option 3: call `__instancecheck__` / `check(...)` over varadic arguments
ins = synonyms('ins', 'instance', 'inschecks')
'''Call the protocol's `__instancecheck__` method over varadic arguments.''';

# Option 4: create a typeguard function from the protocol's `types`
tfn = synonyms('gfn', 'tfn', 'guard', 'typeguard')
'''Use the progocol's `types` to create a typeguard function.''';

# %% ../nbs/01_cons.ipynb 22
class ProtocolMethod(synonyms, Enum):
    '''Methods for the custom protocol type.
    
    Attributes
    ----------
    generator : str
        Try to generate a class from the protocol's `types`.
        
    decorator : str
        Use the protocol as a decorator for a class.
        
    inschecks : str
        Call the protocol's `__instancecheck__` method over varadic arguments.

    typeguard : str
        Use the progocol's `types` to create a typeguard function.
        
    gen : str
        Alias for `generator`.
        
    con : str
        Alias for `generator`.
        
    construct : str
        Alias for `generator`.
    
    dec : str
        Alias for `decorator`.
        
    ins : str
        Alias for `inschecks`.
        
    gfn : str
        Alias for `typeguard`.
        
    tfn : str
        Alias for `typeguard`.
        
    '''
    # Option 1: try to construct a class from the protocol's `types`
    gen: synonyms = gen
    '''Try to generate a class from the protocol's `types`.''';
    
    # Option 2: decorate a class
    dec: synonyms = dec
    
    # Option 3: call `__instancecheck__` / `check(...)` over varadic arguments
    ins: synonyms = ins
    '''Call the protocol's `__instancecheck__` method over varadic arguments.''';
    
    # Option 4: create a typeguard function from the protocol's `types`
    tfn: synonyms = tfn
    '''Use the progocol's `types` to create a typeguard function.''';
    
    @classmethod
    def _missing_(cls: 'ProtocolMethod', val: str, default = None):
        # Case-Insensitive Search
        for mem in cls:
            if mem.ieq(val): return mem
            
        # Check for default sunder method
        if hasattr(cls, _DEFAULT_): 
            edef = cls._default_()
            try: return cls(edef)
            except: ...
            try: return cls(default)
            except: return default
        
        # Check for startswith
        for mem in cls:
            if mem.startswith(val): return mem
            
        return default
    
    def casefold(self) -> str:
        return str(getattr(self, VALUE, NIL)).casefold()
    
    def ieq(self, val: str) -> bool:
        if self.casefold() == str(val).casefold(): return True
        try: 
            if self == str(val).casefold(): return True
        except: return False
    
    def startswith(self, val: str) -> bool:
        if self.casefold().startswith(str(val).casefold()): return True
        try:
            if self.value.startswith(val) == str(val): True
        except: return False
        
    
    @classmethod
    def idgt(cls) -> tuple['ProtocolMethod', ...]:
        return (ProtocolMethod.ins, ProtocolMethod.dec, ProtocolMethod.gen, ProtocolMethod.tfn)
    
    @classmethod
    def gidt(cls) -> tuple['ProtocolMethod', ...]:
        return (ProtocolMethod.gen, ProtocolMethod.dec, ProtocolMethod.ins, ProtocolMethod.tfn)

# %% ../nbs/01_cons.ipynb 24
class ProtocolMethodOrder(Enum):
    idgt = (ProtocolMethod.ins, ProtocolMethod.dec, ProtocolMethod.gen, ProtocolMethod.tfn)
    '''Options for the overloaded `ProtoType.__call__` method.''';
    gidt = (ProtocolMethod.gen, ProtocolMethod.dec, ProtocolMethod.ins, ProtocolMethod.tfn)

# %% ../nbs/01_cons.ipynb 26
PROTOCOL_METHOD_ORDER = (ProtocolMethod.ins, ProtocolMethod.dec, ProtocolMethod.gen, ProtocolMethod.tfn)
'''Options for the overloaded `ProtoType.__call__` method.''';

CALLFNCS = PROTOCOL_METHOD_ORDER



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/01_type.ipynb.

# %% auto 0
__all__ = ['T', 'P', 'Types', 'TypeLike', 'Guard', 'TGuard', 'IterT', 'IterStr', 'IterFunc', 'NanGuard', 'IterGuard',
           'IterTGuard', 'AttrGuards']

# %% ../nbs/01_type.ipynb 6
from math import nan as pynan

# %% ../nbs/01_type.ipynb 8
from types import NoneType, FunctionType
from typing import (
    Self, Type, TypeVar, TypeGuard, TypeAlias, ParamSpec,
    Callable, Iterable, Optional, Protocol, _ProtocolMeta, 
    Any, Set, Dict, List, Tuple, Union, ClassVar, Generic, Literal,
    get_args, get_origin, runtime_checkable, overload, 
)

# %% ../nbs/01_type.ipynb 10
#| export


# %% ../nbs/01_type.ipynb 12
#| export


# %% ../nbs/01_type.ipynb 14
#| export


# %% ../nbs/01_type.ipynb 16
T = TypeVar('T')
P = ParamSpec('P')

Types: TypeAlias = Tuple[Type, ...]
'''`*args` type''';

TypeLike: TypeAlias = Union[type, Type, Type[T], TypeAlias]
'''type, Type, type of type, TypeAlias, or `*args` type''';

Guard: TypeAlias = TypeGuard[TypeLike]
'''Generic TypeGuard''';


TGuard: TypeAlias = TypeGuard[T]

IterT: TypeAlias = Iterable[T]
IterStr: TypeAlias = Iterable[str]
IterFunc: TypeAlias = Iterable[FunctionType]
NanGuard: TypeAlias = TypeGuard[pynan]
IterGuard: TypeAlias = TypeGuard[Iterable]
IterTGuard: TypeAlias = TypeGuard[Iterable[T]]
AttrGuards: TypeAlias = Dict[str, Callable[[T, Any], bool]]



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/03_atyp.ipynb.

# %% auto 0
__all__ = ['T', 'O', 'G', 'P', 'Types', 'Guard', 'IterT', 'IterStr', 'IterFunc', 'IterType', 'GuardIter', 'GuardIterT',
           'GuardIterType', 'NanGuard', 'GuardLike', 'AttrGuards', 'OriginTypes', 'OriginGuard', 'HasIterElements',
           'HasHash', 'IsNotSpecial', 'Namespace', 'ImportedAsset']

# %% ../nbs/03_atyp.ipynb 6
import math

# %% ../nbs/03_atyp.ipynb 8
from types import (ModuleType, FunctionType, SimpleNamespace, BuiltinFunctionType, BuiltinMethodType)

import typing
from typing import (Dict, Type, Tuple, Union, TypeVar, TypeAlias, TypeGuard, ParamSpec, Callable, Iterable, )

# %% ../nbs/03_atyp.ipynb 10
#| export


# %% ../nbs/03_atyp.ipynb 12
from nlit import __HASH__

# %% ../nbs/03_atyp.ipynb 14
try: from ptyp.duck import NonStrBytesIterable, Hashable, NotSpecial
except: ...

# %% ../nbs/03_atyp.ipynb 17
T = TypeVar('T')
'''Generic type variable that can represent any type.''';

O = TypeVar('O')
'''Generic type variable that can represent any object.''';

G = TypeVar('G')
'''Generic type variable that can represent any Graph.''';

P = ParamSpec('P')
'''Parameter specification type variable, capturing the type of function parameters.''';

Types: TypeAlias = Tuple[Type, ...]
'''TypeAlias for a tuple of types, representing a collection of type objects.''';

Guard: TypeAlias = TypeGuard[T]
'''TypeAlias for TypeGuard that specifically guards an object of type T.''';

IterT: TypeAlias = Iterable[T]
'''TypeAlias for any iterable containing elements of type T.''';

IterStr: TypeAlias = Iterable[str]
'''TypeAlias for an iterable of strings.''';

IterFunc: TypeAlias = Iterable[FunctionType]
'''TypeAlias for an iterable of function objects.''';

IterType: TypeAlias = Iterable[Type]
'''TypeAlias for an iterable of types.''';

GuardIter: TypeAlias = TypeGuard[Iterable]
'''TypeAlias for TypeGuard specifically guarding any iterable object.''';

GuardIterT: TypeAlias = TypeGuard[Iterable[T]]
'''TypeAlias for TypeGuard, guarding iterables containing elements of type T.''';

GuardIterType: TypeAlias = TypeGuard[Iterable[Type]]
'''TypeAlias for TypeGuard, guarding iterables containing elements of types.''';

NanGuard: TypeAlias = TypeGuard[math.nan]
'''TypeAlias for TypeGuard specifically guarding the math.nan object.''';

GuardLike: TypeAlias = Callable[[O], TypeGuard[T]]
'''TypeAlias for a callable that takes an object of type O and returns a TypeGuard for type T.''';

AttrGuards: TypeAlias = Dict[str, Callable[[O, T], TypeGuard[T]]]
'''TypeAlias for a dictionary mapping attribute names to guard functions, 
which take an object of type T and an arbitrary value and return a boolean.''';

# %% ../nbs/03_atyp.ipynb 19
OriginTypes: TypeAlias = Union[
    (
        otypes := [
            otype for attr in dir(typing) 
            if (
                typing.get_origin(otype := getattr(typing, attr)) 
                and hasattr(otype, __HASH__)) and 
                not (
                    isinstance(otype, typing._SpecialForm) or otype in (typing.Generic, typing.Protocol)
                )
        ]
    )[0],
    *otypes[1:]
]
'''A union of types from the `typing` module that:
    - have origins (i.e., they are not base types), 
    - are hashable, and 
    - are not special forms or the `Generic` and `Protocol` types. 

This includes types like `List[int]`, `Dict[str, int]`, etc., but excludes basic types like 
`int`, `str`, and `typing` specific constructs like `Generic` or `Protocol`.'''

OriginGuard: TypeAlias = TypeGuard[OriginTypes]
'''TypeGuard for `OriginTypes`. It is used to narrow down a variable to one of the types that have origins in the `typing` module,
are hashable, and not special forms or `Generic` or `Protocol` types. Useful in type checking scenarios to confirm the type of an object
against these specific origin types from the `typing` module.''';

# %% ../nbs/03_atyp.ipynb 21
HasIterElements: TypeAlias = TypeGuard['NonStrBytesIterable']
'''TypeAlias for TypeGuard specifically guarding an iterable object that is not a string or bytes. 
This is useful to find objects which can be recursed into and each element may not be uniformly of the same type.
''';

HasHash: TypeAlias = TypeGuard['Hashable']
'''TypeAlias for TypeGuard specifically guarding a hashable object.''';

IsNotSpecial: TypeAlias = TypeGuard['NotSpecial']
'''TypeAlias for TypeGuard specifically guarding an object that is not a special form.''';

# %% ../nbs/03_atyp.ipynb 23
Namespace: TypeAlias = Union[Dict[str, T], SimpleNamespace]
'''TypeAlias for a union of a dictionary and a `SimpleNamespace` object.''';

ImportedAsset: TypeAlias = Union[T, Type, FunctionType, BuiltinFunctionType, BuiltinMethodType, ModuleType]
'''TypeAlias for a union of a function, a builtin function, a builtin method, and a module object.''';



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/06_prot.ipynb.

# %% auto 0
__all__ = ['defprops', 'ProtocolBaseMeta', 'ProtocolMeta', 'AliasMeta', 'Alias', 'NotMeta', 'Not', 'OptionalMeta', 'Opt',
           'PrototypeMeta', 'Prototype']

# %% ../nbs/06_prot.ipynb 6
from abc import abstractmethod
from functools import wraps

# %% ../nbs/06_prot.ipynb 8
from types import (NoneType,)
from typing import (
    Self, Type, Union, TypeGuard, ClassVar, Callable, 
    Iterable, Optional, Protocol, _ProtocolMeta, runtime_checkable, overload, 
)

# %% ../nbs/06_prot.ipynb 10
#| export


# %% ../nbs/06_prot.ipynb 12
from nchr import NIL, TILDE
from nlit import (
    ALT,  TYPES, PREFIX, SUFFIX, CLSNAME, DROPNONE, 
    __QUALNAME__, __PROPS__, __MODULE__, __ANNOTATIONS__, __BASES__, 
)
from pdec import Property, slotprops
from matr import matr

# %% ../nbs/06_prot.ipynb 14
from .cons import (ALLATTRS, CALLFNCS, ProtocolMethod)
from .atyp import T, Types, IterT, IterFunc, IterStr, Guard, AttrGuards
from .grds import isalias
from .util import opttypes, filtkwds, qualname, getname

# %% ../nbs/06_prot.ipynb 16
defprops = dict(
    btype    = Property('btype', type, None, 'basetype for the protocol'),
    types    = Property('types', IterT, tuple(), 'types that the protocol represents'),
    guards   = Property('guards', IterFunc, tuple(), 'guard functions used to validate the protocol'),
    attrs    = Property('attrs', dict, dict(), 'attributes and their expected values for instances of the protocol'),
    hasattrs = Property('hasattrs', IterStr, tuple(), 'attributes that instances of the protocol must have'),
    allattrs = Property('allattrs', IterStr, ALLATTRS, 'attrs of which all values in the iterable are to be checked against the type guards.'),
    mapattrs = Property('mapattrs', AttrGuards, dict(), 'attributes and custom functions to be used to validate the protocol'),
    priority = Property('priority', IterStr, CALLFNCS, 'priority of the overloaded __call__ method'),
    inverted = Property('inverted', bool, False, 'whether the protocol is inverted'),
    optional = Property('optional', bool, False, 'whether the protocol is optional'),
)
'''default properties for protocols''';

# %% ../nbs/06_prot.ipynb 17
@slotprops(defprops)
class ProtocolBaseMeta(_ProtocolMeta):
    '''A metaclass for creating protocol base classes with custom attributes and guards.

    This metaclass allows for the definition of custom protocol classes with specific type checking,
    attribute requirements, and guard functions.

    Attributes
    ----------
    types : IterT
        Types to check against in the protocol.
        
    guards : IterFunc
        Additional guard functions for type checking.
        
    attrs : dict
        Attributes to compare values against.
        
    hasattrs : IterStr
        Attributes that must exist in the instance.
        
    allattrs : IterStr
        Attributes to check that values are all the same if the object is iterable.
        
    mapattrs : AttrGuards
        Attributes to check with custom functions.
        
    priority : IterStr
        Priority of the overloaded `__call__` method.
        
    inverted : bool
        Whether the protocol is inverted.
        
    optional : bool
        Whether the protocol is optional.
        
    _slots : ClassVar[dict[str, Property]]
        Defined in `slotprops` decorator.

    Methods
    -------
    __init__(self, name, bases, dct, **kwargs)
        Initialize the ProtocolBaseMeta instance.
        
    __new__(mcls, name, bases, dct, **kwargs)
        Create a new class instance.
        
    __instanceguard__(self, ins)
        Check if an instance conforms to the protocol.
        
    guard(self, obj)
        The actual guard function for the protocol.
        
    __instancecheck__(self, ins)
        Check if an instance conforms to the protocol.
        
    __invopt__(self)
        Invert optional.
        
    __invert__(self)
        Invert the protocol.
        
    __not__(self)
        Negate the protocol.
        
    default_kwds(cls, **kwargs)
        Return default keyword arguments.
        
    setqualname(cls, *types, name, **kwargs)
        Set the qualified name of the class.
        
    getbtype(cls)
        Base type (subclass of protocol) that the metaclass should reference.
        
    makebases(cls, kls)
        Determine the base classes for a new class instance.
        
    check(self, *vals)
        Map `cls.guard` to a tuple of values.
        
    new(cls, *types, **kwargs)
        Use class as a decorator.
        
    construct(cls, *args, **kwargs)
        Construct a class from `cls.types`.
        
    typeguard(cls, *args, **kwargs)
        Construct a typeguard for `cls.types`.
        
    __redirect__(cls, *args, __generator, __decorator, __inschecks, **kwargs)
        Redirect to a specific method based on keyword arguments.
        
    __call__(cls, *args, **kwargs)
        Overloaded call method for various functionalities.

    Examples
    --------    
    >>> class ProtocolMeta(ProtocolBaseMeta):
    ...     \'\'\'A metaclass for creating protocol classes with custom type checking.
    ...
    ...     Methods
    ...     -------
    ...     getbtype(cls)
    ...         Returns the base type for the protocol class.
    ...    \'\'\'
    >>>    @classmethod
    >>>    def getbtype(cls): return ProtoType
    ...
    >>> @runtime_checkable
    >>> class ProtoType(Protocol):
    >>>    \'\'\'Base protocol class with custom type checking capabilities.\'\'\'
    '''
    types: IterT = tuple()        # types to check
    guards: IterFunc = tuple()    # additional guards to check
    attrs: dict = dict()          # attributes compare values against
    hasattrs: IterStr = tuple()   # attributes to ensure exist
    allattrs: IterStr = ALLATTRS  # attributes to check that values are all the same if object is iterable
    mapattrs: AttrGuards = dict() # attributes to check with custom functions
    priority: IterStr = CALLFNCS  # priority of the overloaded __call__ method
    inverted: bool = False        # whether the protocol is inverted
    optional: bool = False        # whether the protocol is optional
    __PROPS__: ClassVar[dict[str, Property]] # defined in `slotprops` decorator (was formally ._slots)
    # clsattrs: ClassVar[tuple[str]] = ('types', 'guards', 'attrs', 'hasattrs', 'allattrs', 'mapattrs', 'priority')
    
    def __init__(self: Self, name: str, bases: tuple = tuple(), dct: dict = dict(), **kwargs):
        # super().__init__(name, bases, dct)
        for attr, prop in self.__props__.items():
            aval = kwargs.get(attr, prop.val)
            setattr(self, prop.prv, aval)
            
        self.types = matr.ib(self.types)
        
    def __new__(mcls: Type[Self], name: str, bases: tuple = tuple(), dct: dict = dict(), **kwargs):
        dct.update(**mcls.getprops(**kwargs))
        # dct.update(**mcls.getprops(mcls, **kwargs))
        new = super().__new__(mcls, name, bases, dct)
        for attr, value in kwargs.items():
            setattr(new, attr, value)
        
        kwargs.update(mcls.default_kwds(**kwargs))
        new.types = matr.ib(new.types)
        new.setqualname(*new.types, **kwargs)
        return new

    # @classmethod
    def __init_subclass__(cls: Type[Self], *args, **kwargs):
        super().__init_subclass__(*args, **kwargs)
        for attr, aval in cls.getprops(*args, **kwargs).items():
        # for attr, aval in cls.getprops(cls, *args, **kwargs).items():
            if attr == TYPES:
                setattr(cls, TYPES, opttypes(aval, kwargs.get(DROPNONE, True)))
            else: 
                setattr(cls, attr, aval)
                
    def __invert__(self: Self) -> Self:
        '''Invert the protocol.'''
        self.inverted = not self.inverted
        return self
    
    def __not__(self: Self) -> Self:
        '''Negate the protocol.'''
        self.inverted = not self.inverted
        return self
    
    def __invopt__(self: Self) -> Self:
        '''Invert optional.'''
        self.optional = not self.optional
        return self
                        
    def __instanceguard__(self: Self, ins: object) -> Guard:
        '''Instance method to check if an instance conforms to the protocol.'''
        self.types = matr.ib(self.types)
        res = isalias(ins, self.types, self.guards, self.attrs, self.hasattrs, self.allattrs, self.mapattrs)
        if self.optional: 
            res = res or ins is None
        if self.inverted: 
            res = not res
        return res
    
    def guard(self, obj: object) -> Guard:
        '''The actual guard function for the protocol.'''
        return self.__instanceguard__(obj)

    def __instancecheck__(self: Self, ins: object) -> Guard:
        '''Instance method to check if an instance conforms to the protocol.'''
        return self.guard(ins)
    
    def check(self: Self, *vals) -> Iterable[Guard]:
        '''Map `cls.guard` to a tuple of values.'''
        if len(vals) == 1:
            return self.guard(vals[0])
        return tuple(map(self.guard, vals))

    @classmethod
    def default_kwds(cls: Type[Self], **kwargs) -> dict:
        return {k: kwargs.get(k, v) for k, v in zip((ALT, PREFIX, SUFFIX, DROPNONE), (None, NIL, NIL, True))}

    # @classmethod
    def setqualname(cls: Type[Self], *types, name: str = None, **kwargs) -> Type[Self]: 
        if name is None or name == NIL:
            kwds = filtkwds(qualname, **kwargs)
            kwds.pop(TYPES, None)
            name = qualname(cls, *types, **kwds)
        setattr(cls, __QUALNAME__, name)
        return cls
    
    @classmethod
    @abstractmethod
    def getbtype(cls) -> type:
        '''base type (subclass of protocol) that the metaclass should reference.'''
        return None
    
    @classmethod
    def makebases(cls: Type[Self], kls: Optional[T] = None) -> Types:
        '''Class method to determine the base classes for a new class instance.'''
        btype = cls.getbtype()
        
        nobases = not hasattr(kls, __BASES__) and not isinstance(kls, type)
        if kls is None or nobases: return (btype, )
        bases = (kls, btype) if btype not in kls.__bases__ else (kls, )
        return bases
    
    @classmethod
    def new(cls: Type[Self], *types: Types, **kwargs) -> Type['ProtocolBaseMeta']:
        def decorator(kls: Optional[T] = None):
            clsname = kwargs.get(CLSNAME, getname(kls))
            bases = cls.makebases(kls)
            cdict = dict()
            new = cls.__class__(clsname, bases, cdict, types=types, **kwargs)
            return new
        return decorator
    
    # @classmethod
    def construct(cls: Type[Self], *args, **kwargs) -> Type[Self]:
        '''Try to construct a class from `cls.types`.'''
        for typ in cls.types:
            if callable(typ):
                try: return typ(*args, **kwargs)
                except: continue
        return None
    
    def typeguard(cls: Self) -> Callable[[T], TypeGuard[Self]]:
        f'''Generate the `is{cls.__name__}` type guard method for the protocol.'''
        cname = cls.__name__
        fname = f'is{cname}'
        @wraps(cls.guard, assigned=(__MODULE__, __ANNOTATIONS__))
        def isins(x: object) -> TypeGuard[Self]:
            f'''Check if `x` is an instance of `{cname}`.'''
            return isinstance(x, cls)
        
        isins.__name__ = fname
        isins.__qualname__ = fname
        return isins
    
    def __redirect__(
        cls: Type[Self], *args,
        __generator: bool = False, 
        __decorator: bool = False, 
        __inschecks: bool = False,
        __typeguard: bool = False,
        **kwargs
    ) -> Union[Iterable[Guard], Type[Self], Type['ProtocolBaseMeta']]:
        '''Call a one of the methods invoked by the overloaded `__call__` method by a specific
        keyword argument.
        
        Parameters
        ----------
        args : tuple
            Arguments to pass to the method.
            
        kwargs : dict
            Keyword arguments to pass to the method. 
        
        Other Parameters
        ----------------
        __generator : bool
            Whether to try and construct a class from `cls.types`.
            If True, call `cls.con` method.
            
        __decorator : bool
            Whether to use class as a decorator.
            If True, call `cls.new` method.
            
        __inschecks : bool
            Whether to invoke `__instancecheck__`s. If True, calls 
            `cls.check` method.
            
        __typeguard : bool
            Whether to generate a type guard method. If True, calls
            `cls.typeguard` method.
            
        __gen : bool
            Shorthand alias for `__generator`
            
        __dec : bool
            Shorthand alias for `__decorator`
            
        __ins : bool
            Shorthand alias for `__inschecks`
            
        __tfn : bool
            Shorthand alias for `__typeguard`
        '''
        __generator = __generator or kwargs.pop('__gen', __generator)
        __decorator = __decorator or kwargs.pop('__dec', __decorator)
        __inschecks = __inschecks or kwargs.pop('__ins', __inschecks)
        __inschecks = __inschecks or kwargs.pop('__tfn', __inschecks)
        for opt in cls.priority:
            match opt:
                case ProtocolMethod.ins:
                    if __inschecks: return cls.check(*args)
                case ProtocolMethod.dec:
                    if __decorator: return cls.new(*args, **kwargs)
                case ProtocolMethod.gen:
                    if __generator: return cls.construct(*args, **kwargs)
                case ProtocolMethod.tfn:
                    if __typeguard: return cls.typeguard()
                case _: continue
        return None
    
    @overload
    def __call__(cls: 'ProtocolBaseMeta', base: type, *dtypes: Types, **kwargs) -> Guard: ...
    @overload
    def __call__(cls: type, instance: object) -> Guard: ...
    @overload
    def __call__(cls: type, *instances: tuple[object, ...]) -> Guard: ...
    @overload
    def __call__(cls: type, *args, **kwargs) -> Guard: ...    
    def __call__(cls: Self, *args, **kwargs) -> Guard:
        res = cls.__redirect__(*args, **kwargs)
        if res is not None: return res
        
        # NOTE: if we get here, we will have to guess which method to use
        hasargs = lambda: len(cls.types) > 0 and len(args) > 0
        arg0cls = lambda: isinstance(args[0], type) or args[0] is NoneType
        
        # option one, use the `check` instances
        if hasargs() and not arg0cls():
            gidx = cls.priority.index('gen')
            iidx = cls.priority.index('ins')
            if gidx < iidx: return cls.construct(*args, **kwargs)
            return cls.check(*args)
        
        # option two: create a new class instance via decorator
        deco = cls.new(*args, **kwargs)
        return deco

# %% ../nbs/06_prot.ipynb 18
class ProtocolMeta(ProtocolBaseMeta):
    '''A metaclass for creating protocol classes with custom type checking.

    This metaclass extends `ProtocolBaseMeta` to provide additional functionalities 
    specific to protocol classes.

    Methods
    -------
    getbtype(cls)
        Returns the base type for the protocol class.

    Examples
    --------
    >>> @runtime_checkable
    >>> class ProtoType(Protocol):
    >>>    \'\'\'Base protocol class with custom type checking capabilities.\'\'\'
    '''
    @classmethod
    def getbtype(cls): return Protocol

# %% ../nbs/06_prot.ipynb 20
class AliasMeta(ProtocolMeta):
    '''Metaclass for creating alias classes with custom type checking.'''
    @classmethod
    def getbtype(cls): return Alias

@runtime_checkable
class Alias(Protocol, metaclass=AliasMeta):
    '''Class representing an alias with custom type checking.
    
    Examples
    --------
    >>> @Alias(str, bytes)
    >>> class StrBytes: ...

    >>> @Alias(int, str)
    >>> class IntStr: ...

    >>> IntStr, type(IntStr), IntStr.types
    (abc.IntStr({int, str}), __main__.AliasMeta, (int, str))

    >>> IntStr(1, '2', None, [], __ins=True)
    (True, True, False, False)
    '''

# %% ../nbs/06_prot.ipynb 22
class NotMeta(ProtocolMeta):
    '''Metaclass for creating negation classes with custom type checking.'''
    @classmethod
    def getbtype(cls): return Not

    @classmethod
    def default_kwds(cls, **kwargs) -> dict:
        kwds = super().default_kwds(**kwargs)
        kwds[PREFIX] = TILDE
        return kwds
    
    def guard(self, obj) -> Guard:
        return not super().guard(obj)
    
@runtime_checkable
class Not(Protocol, metaclass=NotMeta):
    '''Class representing a negation with custom type checking.
    
    Examples
    --------
    >>> @Not(str, bytes)
    >>> class NotStrBytes: ...

    >>> @Not(int)
    >>> class NotInt: ...

    >>> vals = (10, None, 'hello', 42.30, tuple())
    >>> NotInt(*vals,  __ins=True)
    (False, True, True, True, True)
    '''

# %% ../nbs/06_prot.ipynb 24
class OptionalMeta(ProtocolMeta):
    '''Metaclass for creating optional type classes with custom type checking.'''
    @classmethod
    def getbtype(cls): return Opt

    @classmethod
    def default_kwds(cls, **kwargs) -> dict:
        kwds = super().default_kwds(**kwargs)
        kwds[PREFIX] = '?'        
        return kwds
    
    def __new__(mcls, name, bases, dct, **kwargs):
        '''Create a new class instance.'''
        types = kwargs.pop(TYPES, tuple())
        for ntype in (None, ):
            if ntype not in types:
                types += (ntype, )
        dct[TYPES] = types
        new = super().__new__(mcls, name, bases, dct, types = types, **kwargs)
        return new
    
@runtime_checkable
class Opt(Protocol, metaclass=OptionalMeta):
    '''Class representing an optional type with custom type checking.
    
    Examples
    --------
    >>> @Opt(str)
    >>> class StrQ: ...

    >>> @Opt(int)
    >>> class IntQ: ...

    >>> vals = (10, None, 'hello', 42.30, tuple())
    >>> IntQ(*vals,  __ins=True)
    (True, True, False, False, False)

    >>> @Alias(IntQ, str)
    >>> class IntQStr: ...

    >>> IntQStr(*vals,  __ins=True)
    (True, True, True, False, False)
    '''

# %% ../nbs/06_prot.ipynb 26
class PrototypeMeta(ProtocolMeta):
    '''Metaclass for creating optional type classes with custom type checking.'''  
    q: Opt # optional version of self (i.e. with `None` added to `types`)
    n: Not # negated version of self (i.e. `not self.guad()`)
    
    @classmethod
    def getbtype(cls): return Prototype
    def __new__(mcls: Type[Self], name: str, bases: tuple = tuple(), dct: dict = dict(), **kwargs):
        types = kwargs.pop(TYPES, ())
        new = super().__new__(mcls, name, bases, dct, types=types, **kwargs)
        
        @Not(new, **kwargs)
        class notcls: ...
        
        @Opt(new, **kwargs)
        class optcls: ...
        
        notcls.__qualname__ = qualname(new, *new.types, **Not.default_kwds(**kwargs))
        optcls.__qualname__ = qualname(new, *new.types, **Opt.default_kwds(**kwargs))
        notcls.__name__ = f'{TILDE}{new.__name__}'
        optcls.__name__ = f'?{new.__name__}'
        
        setattr(new, 'n', notcls)
        setattr(new, 'q', optcls)
        return new      
    
@runtime_checkable
class Prototype(Protocol, metaclass=PrototypeMeta):
    '''Class representing an alias with custom type checking.
    
    Examples
    --------
    >>> @Alias(str, bytes)
    >>> class StrBytes: ...
    ...
    >>> @Alias(int, str)
    >>> class IntStr: ...
    ...
    >>> @Not(str, bytes)
    >>> class NotStrBytes: ...
    ...
    >>> @Not(int)
    >>> class NotInt: ...
    ...
    >>> @Opt(str)
    >>> class StrQ: ...
    ...
    >>> @Opt(int)
    >>> class IntQ: ...
    ...
    >>> @Alias(IntQ, str)
    >>> class IntQStr: ...
    ...
    >>> vals = (10, None, 'hello', 42.30, tuple())
    >>> tuple(map(lambda c: c(*vals, __ins=True), (IntStr, NotInt, IntQ, IntQStr, )))
    ( # (10,    None, 'hello', 42.30,  tuple())
        (True, False,    True, False,  False), # IntStr
        (False, True,    True,  True,   True), # NotInt
        (True,  True,   False, False,  False), # OptInt
        (True,  True,    True, False,  False)  # IntQStr
    )
    
    ...
    ...
    >>> @Typed(int)
    >>> class tint: ...
    ...
    >>> tuple(((k.__name__, k.__qualname__) for k in (tint, tint.n, tint.q)))
    (('tint', 'tint({int})'), ('~tint', 'tint(~{int})'), ('?tint', 'tint(?{int})'))
    ...
    >>> tint(None, 1, 's', __ins=True), tint.q(None, 1, 's', __ins=True), tint.n(None, 1, 's', __ins=True)
    ((False, True, False), (True, True, False), (True, False, True))
    ...
    >>> vals = (10, None, 'hello', 42.30, tuple())
    >>> tuple(map(lambda x: (tint(x, __ins=True), tint.q(x, __ins=True), tint.n(x, __ins=True)), vals))
    # tint, tint.q, tint.n)
    ((True,  True,  False),
    (False, True,   True),
    (False, False,  True),
    (False, False,  True),
    (False, False,  True))
    ...
    ...
    # prioritize the order of the __call__ methods to `('ins', 'gen', 'dec')` 
    # i.e. (`check`, `con`, `new`). Note that the default order is `('gen', 'ins', 'dec')`
    # NOTE: default order of the __call__ methods (`check`, `con`, `new`)
    >>> @Typed(int, priority=('ins', 'gen', 'dec'))
    >>> class tint: ...
    ...
    >>> tint(None, 1, 's'), tint.q(None, 1, 's'), tint.n(None, 1, 's')
    ((False, True, False), (True, True, False), (True, False, True))
    '''
    q: Opt # optional version of self (i.e. with `None` added to `types`)
    n: Not # negated version of self (i.e. `not self.guad()`)



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/05_ptyp.ipynb.

# %% auto 0
__all__ = ['StrBytes', 'NotStrBytes', 'StrQ', 'IntQ']

# %% ../nbs/05_ptyp.ipynb 6
import os, inspect
from abc import abstractmethod
from functools import wraps, singledispatch

# %% ../nbs/05_ptyp.ipynb 8
from types import NoneType 
from typing import (Iterable, TypeGuard, )

# %% ../nbs/05_ptyp.ipynb 10
#| export


# %% ../nbs/05_ptyp.ipynb 12
#| export


# %% ../nbs/05_ptyp.ipynb 14
from .meta import (Alias, Not, Opt)

# %% ../nbs/05_ptyp.ipynb 16
@Alias(str, bytes)
class StrBytes: ...

@Not(str, bytes)
class NotStrBytes: ...

@Opt(str)
class StrQ: ...

@Opt(int)
class IntQ: ...
